/*
 * $Id: IterativeResolver.java 325 2006-05-22 14:59:32Z davidb $
 *
 * Copyright (c) 2004 USC Project SRO. All rights reserved.
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.iter;

import java.util.*;

import org.apache.log4j.Logger;

import org.xbill.DNS.CNAMERecord;
import org.xbill.DNS.DClass;
import org.xbill.DNS.Flags;
import org.xbill.DNS.Name;
import org.xbill.DNS.Rcode;
import org.xbill.DNS.Section;
import org.xbill.DNS.Type;

import se.rfc.unbound.*;
import se.rfc.unbound.cache.DNSCache;
import se.rfc.unbound.cache.HostCache;
import se.rfc.unbound.cache.HybridDNSCache;
import se.rfc.unbound.cache.StandardHostCache;

/**
 * This is a resolver module that implements the Iterative resolver logic:
 * that is, given a request, iteratively query the DNS until the final answer
 * is reached.
 * 
 * @author davidb
 * @version $Revision: 325 $
 */
public class IterativeResolver extends AbstractResolver
{
  /** The main cache. */
  private DNSCache             mDNSCache;

  /** The host cache. */
  private HostCache            mHostCache;

  /**
   * The hints -- these aren't stored in the cache because they don't expire.
   * The hints are always used to "prime" the cache. Note that both root hints
   * and stub zone "hints" are stored in this data structure.
   */
  private Hints                mHints;

  /** A flag to indicate whether or not we have an IPv6 route */
  private boolean              mSupportsIPv6;

  /**
   * A mapping of forwarding zones to targets. This will be null if there are
   * no forwarding zones.
   */
  private Map                  mForwardTable       = null;

  /**
   * This is a map of in-progress, locally generated requests. It is used to
   * prevent duplication of effort.
   */
  private ConcurrentRequestMap mConcurrentRequestMap;

  /** A set of InetAddress objects that should never be queried. */
  private Set                  mDoNotQueryAddresses;

  /** An algorithm for selecting the "best" nameserver target. */
  private ServerSelection      mServerSelection;

  /** The maximum dependency depth that this resolver will pursue. */
  private int                  mMaxDependencyDepth = 4;

  /**
   * The target fetch policy for each dependency level. This is described as a
   * simple number (per dependency level): negative numbers (usually just -1)
   * mean fetch-all, 0 means only fetch on demand, and positive numbers mean
   * to fetch at most that many targets.
   */
  private int[]                mTargetFetchPolicy;

  /**
   * The TTL to use for the "private TTL" feature. If zero, this feature will
   * not be used
   */
  private long                 mPrivateTTL         = 0;

  private Logger               log                 = Logger.getLogger(this
                                                       .getClass());

  /*
   * This module needs to follow a few basic rules.
   * 
   * One, it must take care not to send the exact same DNSEvent instance to
   * the next module (NSClient) more than once at a time. That is, if that
   * event has been sent to NSClient, do not send it again until we get the
   * response to it from NSClient. The reason for this is that NSClient will
   * treat every DNSEvent as unique, even if it is already processing the
   * exact same instance object.
   */

  public IterativeResolver(Resolver nextResolver,
      ServerSelection serverSelection)
  {
    super(nextResolver);
    mServerSelection = serverSelection;
    mConcurrentRequestMap = new ConcurrentRequestMap();
  }

  public void init(Properties config) throws Exception
  {
    super.init(config);

    mHints = new Hints();

    // instantiate our DNSCache.

    String s = config.getProperty("dns.cache.class");
    if (s == null)
    {
      // Default cache implementation
      mDNSCache = new HybridDNSCache();
    }
    else
    {
      mDNSCache = IterUtils.loadDNSCache(s);
    }

    mDNSCache.init(config);

    mHostCache = new StandardHostCache();
    mHostCache.init(config);

    // FIXME: for now, we can only handle one set of root hints (other
    // sets would be for other classes).
    s = config.getProperty("dns.root_hints_file");
    mHints.configureRootHints(s);

    // Load forwarding instructions
    mForwardTable = IterUtils.configureForwardZones(config);

    // Load stub zone hints.
    s = config.getProperty("dns.stub");
    if (s != null)
    {
      mHints.configureStubZones("dns.stub", config);
    }

    // Load the "private TTL" parameter.
    s = config.getProperty("dns.private_ttl");
    mPrivateTTL = Util.parseLong(s, 0);
    if (mPrivateTTL > 0)
    {
      log.debug("Private TTL is " + mPrivateTTL);
    }

    // Load any forbidden query addresses.
    // The default is to use the server.address as a forbidden address.
    s = config.getProperty("dns.noquery.addresses");
    if (s != null)
    {
      mDoNotQueryAddresses = IterUtils.parseAddressSet(s);
    }
    else if ((s = config.getProperty("server.address")) != null)
    {
      mDoNotQueryAddresses = IterUtils.parseAddressSet(s);
    }

    // Load our max dependency depth, if one was defined
    s = config.getProperty("dns.max_dependency_depth");
    if (s != null)
    {
      mMaxDependencyDepth = Util.parseInt(s, mMaxDependencyDepth);
      log.debug("Maximum dependency depth: " + mMaxDependencyDepth);
    }

    // Load the per-dependency-depth target fetch policy
    s = config.getProperty("dns.targetfetch.policy");
    mTargetFetchPolicy = IterUtils
        .parseTargetFetchPolicy(mMaxDependencyDepth, s);
    for (int i = 0; i < mTargetFetchPolicy.length; i++)
    {
      log.debug("Target fetch policy for level " + i + ": "
          + mTargetFetchPolicy[i]);
    }
  }

  /**
   * This is a convenience routine for fetching the state object from a given
   * event.
   * 
   * @param event The event in question.
   * @return The state object for this module, if there is one, null if not.
   */
  private IterEventState getModuleState(DNSEvent event)
  {
    IterEventState state = (IterEventState) event.getModuleState(this);
    return state;
  }

  /**
   * Generate a local request event. Local events are tied to this module, and
   * have a correponding (first tier) event that is waiting for this event to
   * resolve to continue.
   * 
   * @param qname The query name for this request.
   * @param qtype The query type for this request.
   * @param qclass The query class for this request.
   * @param forEvent The event that is generating this event.
   * @param initial_request_state The next state for the request.
   * @param initial_response_state The initial response state (normally this
   *          is QUERY_RESP_STATE, unless it is known that the request won't
   *          need iterative processing).
   * @param final_response_state The final state for the response to this
   *          request.
   * 
   * @return The generated event.
   */
  private DNSEvent generateLocalRequest(Name qname, int qtype, int qclass,
      DNSEvent forEvent, int initial_state, int final_state)
  {
    Request r = new Request(qname, qtype, qclass);
    // RD should be set only when sending the query back through the INIT
    // state.
    r.setFlag(Flags.RD, initial_state == IterEventState.INIT_REQUEST_STATE);

    // We set the CD flag so we can send this through the "head" of the
    // resolution chain, which might have a validator. We are uninterested in
    // validating things not on the direct resolution path.
    r.setFlag(Flags.CD);

    DNSEvent event = new DNSEvent(r, forEvent);
    IterEventState state = new IterEventState();

    state.state = initial_state;
    state.finalState = final_state;

    event.setModuleState(this, state);

    return event;
  }

  /**
   * Given a response event, remove suspect RRsets from the response.
   * "Suspect" rrsets are potentially poison. Note that this routine expects
   * the response to be in a "normalized" state -- that is, all "irrelevant"
   * RRsets have already been removed, CNAMEs are in order, etc.
   * 
   * @param event The response event.
   * @param request The request for this event (already extracted).
   * @param state The local state for this event (already extracted).
   */
  private void sanitize(DNSEvent event, Request request, IterEventState state)
  {
    Response response = event.getResponse();
    if (response == null) return;

    // From the request, we get the "originating zone".
    DelegationPoint dp = state.currentDelegationPoint;
    if (dp == null) return;

    Name originatingZone = dp.getName();

    SMessage m = response.getSMessage();

    // At this point, we brutally remove ALL rrsets that aren't children of
    // the originating zone. The idea here is that, as far as we know, the
    // server that we contacted is ONLY authoritative for the originating
    // zone. It, of course, MAY be authoriative for any other zones, and of
    // course, MAY NOT be authoritative for some subdomains of the originating
    // zone.
    
    for (int section = Section.ANSWER; section <= Section.ADDITIONAL; section++)
    {
      List rrset_list = m.getSectionList(section);
      for (ListIterator i = rrset_list.listIterator(); i.hasNext();)
      {
        SRRset rrset = (SRRset) i.next();
        int type = rrset.getType();
        
        // skip DNAME records -- they will always be followed by a synthesized
        // CNAME, which will be relevant.
        // FIXME: should this do something differently with DNAME rrsets NOT
        // in Section.ANSWER?
        if (type == Type.DNAME) continue;
        
        // Otherwise the record must be a subdomain of the originating domain.
        if (!rrset.getName().subdomain(originatingZone))
        {
          log.trace("sanitize: removing potential poison RRset from section " + 
              Section.string(section) + ": " + rrset);
          i.remove();
        }
      }
    }
  }

  /**
   * Given a locally generated request, send it along to be resolved. This
   * will also make the duplicate query check.
   * 
   * @param event The locally generated request event.
   */
  private void processLocalRequest(DNSEvent event)
  {
    // If the add method returns true, then the locally generated query was a
    // duplicate of a different locally generated query. In this case, we
    // don't process both, but just attach this event to the already
    // in-progress one.
    int res = mConcurrentRequestMap.add(event, this);
    switch (res)
    {
      case ConcurrentRequestMap.DUPLICATE :
        log.trace("attaching local request to currently "
            + "in-progress request: " + event);
        return;
      case ConcurrentRequestMap.DEADLOCK :
        log.trace("dependency loop detected. "
            + "failing new local request immediately");
        IterEventState state = getModuleState(event);
        errorResponse(event, event.getRequest(), state, Rcode.SERVFAIL);
        handleEvent(event, event.getRequest(), state);
        return;
    }

    // Otherwise, this is a normal, unduplicated query, so inject it.
    mHeadResolver.processRequest(event);
  }

  /**
   * Given a CNAME response (defined as a response containing a CNAME or DNAME
   * that does not answer the request), process the response, modifying the
   * state as necessary. This follows the CNAME/DNAME chain and returns the
   * final query name.
   * 
   * @param state The event state to modify.
   * @param request The request that generated the response.
   * @param response The cname response itself.
   * 
   * @return The new query name, after following the CNAME/DNAME chain.
   */
  private Name handleCNAMEResponse(IterEventState state, Request request,
      SMessage response)
  {
    // Allocate a list if one hasn't been already. (Lazy allocation)
    if (state.prependList == null)
    {
      state.prependList = new LinkedList();
    }

    // Start with the (current) qname.
    Name sname = request.getQName();

    // Iterate over the ANSWER rrsets in order, looking for CNAMEs and DNAMES.
    SRRset[] rrsets = response.getSectionRRsets(Section.ANSWER);
    for (int i = 0; i < rrsets.length; i++)
    {
      // If there is a (relevant) DNAME, add it to the list.
      // We always expect there to be CNAME that was generated by this DNAME
      // following, so we don't process the DNAME directly.
      if (rrsets[i].getType() == Type.DNAME
          && sname.strictSubdomain(rrsets[i].getName()))
      {
        state.prependList.add(rrsets[i]);
        continue;
      }

      if (rrsets[i].getType() == Type.CNAME
          && rrsets[i].getName().equals(sname))
      {
        // Add this relevant CNAME rrset to the prepend list.
        state.prependList.add(rrsets[i]);

        // We make a good assumption that a CNAME rrset only has one record
        // (not counting RRSIGS).
        // We process the CNAME to find the new name to look for.
        CNAMERecord cname = (CNAMERecord) rrsets[i].first();
        sname = cname.getTarget();
      }

      // Other rrsets in the section are ignored.
    }

    return sname;
  }

  /**
   * Attempt to forward a request. If this module is configured to do so, it
   * will "forward" requests that are at or below certain names. When
   * forwarding, this module temporarly acts like a "stub" resolver in the
   * sense that it does not iteratively/recursively resolve in this state.
   * 
   * @param event The event associated with the request.
   * @param state The state associated with the event & request.
   * @param req The request to possibly forward.
   * 
   * @return true if the request was forwarded, false if not.
   */
  private boolean forwardRequest(DNSEvent event, IterEventState state,
      Request req)
  {
    if (mForwardTable == null) return false;

    // Iterate over the (hopefully short) forward table and see if our current
    // qname is a subdomain of any of the forward zones.
    Name qname = req.getQName();
    Host[] targets = null;
    Name forwardZone = null;
    for (Iterator i = mForwardTable.entrySet().iterator(); i.hasNext();)
    {
      Map.Entry entry = (Map.Entry) i.next();
      forwardZone = (Name) entry.getKey();
      if (qname.subdomain(forwardZone))
      {
        targets = (Host[]) entry.getValue();
        break;
      }
    }

    // If the current qname isn't being forwarded, return.
    if (targets == null || targets.length == 0) return false;

    // Set the RD flag on our to-be-forwarded query.
    req.setFlag(Flags.RD);
    state.currentDelegationPoint = new DelegationPoint(forwardZone, Arrays
        .asList(targets));
    state.targetQueries = 0;

    log.trace("forwarding request to " + state.currentDelegationPoint);
    return true;
  }

  /**
   * Generate and send a root priming request.
   * 
   * @param forEvent the DNSEvent that triggered the need to prime.
   * @param forState the state attached to the forEvent.
   * @param qclass the class to prime.
   * 
   * @return true if the priming request was actually sent, false if not. The
   *         priming request won't be sent if, for instance, there aren't any
   *         root hints for the given class.
   */
  private boolean primeRoot(DNSEvent forEvent, IterEventState forState,
      int qclass)
  {
    log.debug("priming <./" + DClass.string(qclass) + ">");

    SMessage m = mHints.lookupRootHint(qclass);

    if (m == null)
    {
      log.debug("Cannot prime class " + DClass.string(qclass)
          + " due to lack of hints.");
      return false;
    }

    // Priming requests start at the QUERYTARGETS state, skipping the normal
    // INIT state logic (which would cause an infloop).
    DNSEvent priming_request = generateLocalRequest(Name.root,
        Type.NS,
        qclass,
        forEvent,
        IterEventState.QUERYTARGETS_STATE,
        IterEventState.PRIME_RESP_STATE);

    // Set the initial delegation point to the hint.
    IterEventState priming_state = getModuleState(priming_request);
    priming_state.currentDelegationPoint = DelegationPoint
        .messageToDelegationPoint(m, mHostCache);
    priming_state.targetQueries = 0; // suppress any target queries.

    // and send the query.
    processLocalRequest(priming_request);
    return true;
  }

  /**
   * Generate and process a stub priming request. This method tests for the
   * need to prime a stub zone, so it is safe to call for every request.
   * 
   * @param forEvent The DNSEvent that triggered this priming request.
   * @param forState The state associated with the forEvent.
   * @param qname The triggering request's qname.
   * @param qclass the triggering request's qclass.
   * 
   * @return true if a priming request was made, false if not. The will only
   *         issue a priming request if it detects an unprimed stub.
   */
  private boolean primeStub(DNSEvent forEvent, IterEventState forState,
      Name qname, int qclass)
  {
    // Lookup the stub hint. This will return null if the stub doesn't need to
    // be re-primed.
    DelegationPoint stub_dp = mHints.lookupStubHint(qname,
        qclass,
        forState.currentDelegationPoint);

    // The stub (if there is one) does not need priming.
    if (stub_dp == null) return false;

    // Otherwise, we need to (re)prime the stub.

    log.debug("priming stub <" + stub_dp.getName() + "/"
        + DClass.string(qclass) + ">");

    // Stub priming events start at the QUERYTARGETS state to avoid the
    // redundant INIT state processing.
    DNSEvent priming_request = generateLocalRequest(stub_dp.getName(),
        Type.NS,
        qclass,
        forEvent,
        IterEventState.QUERYTARGETS_STATE,
        IterEventState.PRIME_RESP_STATE);

    // Set the initial delegation point to the hint.
    IterEventState priming_state = getModuleState(priming_request);
    priming_state.currentDelegationPoint = stub_dp;
    // suppress target queries -- although there wouldn't be anyway, since
    // stub hints never have missing targets.
    priming_state.targetQueries = 0;
    priming_state.primingStub = true;

    // and send the query.
    processLocalRequest(priming_request);
    return true;
  }

  /**
   * Transition to the next state. This can be used to advance a currently
   * processing event. It cannot be used to reactivate a forEvent.
   * 
   * @param event The event to process.
   * @param req The current request for the event.
   * @param state The module state for the event.
   * @param nextState The state to transition to.
   * @return true. This is so this can be called as the return value for the
   *         actual process*State() methods. (Transitioning to the next state
   *         implies further processing).
   */
  private boolean nextState(DNSEvent event, Request req,
      IterEventState state, int nextState)
  {
    // If transitioning to a "response" state, make sure that there is a
    // response.
    if (IterEventState.isResponseState(nextState))
    {
      if (event.getResponse() == null)
      {
        log.error("transitioning to response state sans response.");
      }
    }

    state.state = nextState;

    return true;
  }

  /**
   * (Re)activate an event. This will cause the event loop to process this
   * event, after setting the event to the nextState.
   * 
   * @param event The event to activate.
   * @param state The module state for that event.
   * @param nextState The state to promote this event to.
   * @return false, so it can be used as the return value for the
   *         processState() methods.
   */
  private boolean activateEvent(DNSEvent event, IterEventState state,
      int nextState)
  {
    state.state = nextState;
    handleEvent(event, event.getRequest(), state);
    return false;
  }

  /**
   * Transition an event to its final state. Final states always either return
   * a result up the module chain, or reactivate a dependent event. Which
   * final state to transtion to is set in the module state for the event when
   * it was created, and depends on the original purpose of the event.
   * 
   * @param event The event to transition.
   * @param req The request for that event.
   * @param state The module state for that event.
   * @return false. This is so this method can be used as the return value for
   *         the processState methods. (Transitioning to the final state
   */
  private boolean finalState(DNSEvent event, Request req, IterEventState state)
  {
    return nextState(event, req, state, state.finalState);
  }

  /**
   * Generate an error response and transition the event to its final state.
   * 
   * @param event The event to transition.
   * @param req The request for that event.
   * @param state The module state for that event.
   * @param rcode The Rcode to use for the error.
   * @return false.
   */
  private boolean errorResponse(DNSEvent event, Request req,
      IterEventState state, int rcode)
  {
    event.setResponse(Util.errorMessage(req, rcode));
    event.getResponse().setResponseType(ResponseType.THROWAWAY);
    return finalState(event, req, state);
  }

  /**
   * Generate an error response to an inactive event, activate the event, and
   * transition it to its final state.
   * 
   * @param event The event to activate.
   * @param state The module state for that event.
   * @param rcode The Rcode to use for the error.
   * @return false.
   */
  private boolean activateErrorResponse(DNSEvent event, IterEventState state,
      int rcode)
  {
    event.setResponse(Util.errorMessage(event.getRequest(), rcode));
    event.getResponse().setResponseType(ResponseType.THROWAWAY);
    activateEvent(event, state, state.finalState);
    return false;
  }

  /**
   * This is the primary entry point for processing request events. Note that
   * this method should only be used by external modules.
   * 
   * @param event The event to process.
   */
  public void processRequest(DNSEvent event)
  {
    // Make sure that this event has a module state.
    IterEventState state = getModuleState(event);
    if (state == null)
    {
      state = new IterEventState();
      event.setModuleState(this, state);
      // external requests start in the INIT state, and finish using the
      // FINISHED state.
      state.state = IterEventState.INIT_REQUEST_STATE;
      state.finalState = IterEventState.FINISHED_STATE;
    }

    log.trace("processRequest: new external request event: " + event);
    handleEvent(event, event.getRequest(), state);
  }

  /**
   * This is the primary entry point for response processing.
   * 
   * @param event The event to process.
   */
  public void processResponse(DNSEvent event)
  {
    IterEventState state = getModuleState(event);

    Request req = event.getRequest();

    log.trace("processResponse: new external response event: " + event);
    state.state = IterEventState.QUERY_RESP_STATE;
    handleEvent(event, req, state);

    if (state.hasConcurrentEvents())
    {
      // copy the concurrent events into a local list to avoid concurrent
      // modification of the list.
      List concurrentEvents = new ArrayList(state.getConcurrentEvents()
          .size());
      concurrentEvents.addAll(state.getConcurrentEvents());
      state.clearConcurrentEvents();

      for (Iterator i = concurrentEvents.iterator(); i.hasNext();)
      {
        DNSEvent c_event = (DNSEvent) i.next();
        Request c_req = c_event.getRequest();
        IterEventState c_state = getModuleState(c_event);
        c_event.setResponse(event.getResponse());

        log.trace("activating attached concurrent event: " + c_event);
        c_state.state = IterEventState.QUERY_RESP_STATE;
        handleEvent(c_event, c_req, c_state);
      }
    }

    mConcurrentRequestMap.remove(event);
  }

  /**
   * Handle events. This is the real processing loop for events, responsible
   * for moving events through the various states. If a processing method
   * returns true, then it will be advanced to the next state. If false, then
   * processing will stop.
   * 
   * @param event The response event.
   * @param req The current request associated with the event.
   * @param state The state associated with the event.
   */
  private void handleEvent(DNSEvent event, Request req, IterEventState state)
  {
    boolean cont = true;

    while (cont)
    {
      log.trace("handleEvent processing: " + event + " with state "
          + IterEventState.stateString(state.state));
      switch (state.state)
      {
        case IterEventState.INIT_REQUEST_STATE :
          cont = processInitRequest(event, req, state);
          break;
        case IterEventState.INIT_REQUEST_2_STATE :
          cont = processInitRequest2(event, req, state);
          break;
        case IterEventState.INIT_REQUEST_3_STATE :
          cont = processInitRequest3(event, req, state);
          break;
        case IterEventState.QUERYTARGETS_STATE :
          cont = processQueryTargets(event, req, state);
          break;
        case IterEventState.QUERY_RESP_STATE :
          cont = processQueryResponse(event, req, state);
          break;
        case IterEventState.PRIME_RESP_STATE :
          cont = processPrimeResponse(event, req, state);
          break;
        case IterEventState.TARGET_RESP_STATE :
          cont = processTargetResponse(event, req, state);
          break;
        case IterEventState.FINISHED_STATE :
          cont = processFinished(event, req, state);
          break;
        default :
          log.warn("Invalid state: "
              + IterEventState.stateString(state.state));
          cont = false;
          break;
      }
    }
  }

  /**
   * Process the initial part of the request handling. This state roughly
   * corresponds to resolver algorithms steps 1 (find answer in cache) and 2
   * (find the best servers to ask).
   * 
   * Note that all requests start here, and query restarts revisit this state.
   * 
   * This state either generates: 1) a response, from cache or error, 2) a
   * priming event, or 3) forwards the request to the next state (init2,
   * generally).
   * 
   * @param event The (request) event being processed.
   * @param req The (current) request for that event.
   * @param state The state for that event.
   * @return true if the event needs more request processing immediately,
   *         false if not.
   */
  private boolean processInitRequest(DNSEvent event, Request req,
      IterEventState state)
  {
    log.trace("resolving <" + req.getQName() + "/"
        + Type.string(req.getQType()) + "/" + DClass.string(req.getQClass())
        + ">");

    // Check effort

    // We enforce a maximum number of query restarts. This is primarily a
    // cheap way to prevent CNAME loops.
    if (state.queryRestartCount > 8)
    {
      log.debug("request has exceeded the "
          + "maximum number of query restarts with "
          + state.queryRestartCount);
      return errorResponse(event, req, state, Rcode.SERVFAIL);
    }

    // We enforce a maximum recursion/dependency depth -- in general, this is
    // unnecessary for dependency loops (although it will catch those), but it
    // provides a sensible limit to the amount of work required to answer a
    // given query.
    log.trace("request has dependency depth of " + event.getDepth());
    if (event.getDepth() > mMaxDependencyDepth)
    {
      log.debug("request has exceeded the maximum dependency depth "
          + "with depth of" + event.getDepth());
      return errorResponse(event, req, state, Rcode.SERVFAIL);
    }

    // Resolver Algorithm Step 1 -- Look for the answer in local data.

    // This either results in a query restart (CNAME cache response), a
    // terminating response (ANSWER), or a cache miss (null).

    SMessage resp = mDNSCache.lookup(req.getQName(), req.getQType(), req
        .getQClass());

    // Handle a positive cache response.
    if (resp != null)
    {
      int type = ResponseType.typeCacheResponse(resp, req);

      // If we have a CNAME response (i.e., not an answer), process it and
      // restart.
      if (type == ResponseType.CNAME)
      {
        log.trace("returning CNAME response from cache: " + resp);
        Name cname = handleCNAMEResponse(state, req, resp);
        req.setQName(cname);
        // This *is* a query restart, even if it is a cheap one.
        state.queryRestartCount++;

        return nextState(event, req, state, IterEventState.INIT_REQUEST_STATE);
      }

      // Otherwise, the cache response is an ANSWER, so finish.
      log.trace("returning answer from cache.");
      event.setResponse(resp);

      return finalState(event, req, state);
    }

    // Attempt to forward the request.
    if (forwardRequest(event, state, req))
    {
      // the request has been forwarded.
      // forwarded requests need to be immediately sent to the next state,
      // QUERYTARGETS.
      return nextState(event, req, state, IterEventState.QUERYTARGETS_STATE);
    }

    // Attempt to find a covering DNAME in the cache
    resp = mDNSCache.findDNAME(req.getQName(), req.getQType(), req
        .getQClass());
    if (resp != null)
    {
      log.trace("returning synthesized CNAME response from cache: " + resp);
      Name cname = handleCNAMEResponse(state, req, resp);
      // At this point, we just initiate the query restart.
      // This might not be a query restart situation (e.g., qtype == CNAME),
      // but
      // the answer returned from findDNAME() is likely to be one that we
      // don't want to return.
      // Thus we allow the cache and other resolution mojo kick in regardless.
      req.setQName(cname);
      state.queryRestartCount++;

      return nextState(event, req, state, IterEventState.INIT_REQUEST_STATE);
    }

    // Resolver Algorithm Step 2 -- find the "best" servers.

    // first, adjust for DS queries. To avoid the grandparent problem, we just
    // look for the closest set of server to the parent of qname.
    Name delname = req.getQName();
    if (req.getQType() == Type.DS)
    {
      delname = new Name(delname, 1);
    }

    // Lookup the delegation in the cache. This will either return a REFERRAL
    // or null. If null, then the cache must be primed for the qclass.
    state.currentDelegationMessage = mDNSCache.findDelegation(delname, req
        .getQClass());

    // If the cache has returned nothing, then we have a root priming
    // situation.
    if (state.currentDelegationMessage == null)
    {
      // Note that the result of this will set a new
      // currentDelegationMessage/Point based on the result of priming.
      if (!primeRoot(event, state, req.getQClass()))
      {
        return errorResponse(event, req, state, Rcode.REFUSED);
      }

      // priming creates an sends a subordinate query, with this query as the
      // forEvent. So further processing for this event will stop until
      // reactivated by the results of priming.
      return false;
    }

    // Reset the RD flag. If this is a query restart, then the RD will have
    // been turned off.
    req.setFlag(Flags.RD, event.getOrigRequest().getFlag(Flags.RD));

    // Otherwise, set the current delegation point and move on to the next
    // state.
    state.currentDelegationPoint = DelegationPoint
        .messageToDelegationPoint(state.currentDelegationMessage, mHostCache);
    return nextState(event, req, state, IterEventState.INIT_REQUEST_2_STATE);
  }

  /**
   * Process the second part of the initial request handling. This state
   * basically exists so that queries that generate root priming events have
   * the same init processing as ones that do not. Request events that reach
   * this state must have a valid currentDelegationMessage and Point set.
   * 
   * This part is primarly handling stub zone priming. Events that reach this
   * state must have a current delegation point and message.
   * 
   * @param event The event being processed.
   * @param req The current request for the event.
   * @param state The state for the event.
   * @return true if the event needs more request processing.
   */
  private boolean processInitRequest2(DNSEvent event, Request req,
      IterEventState state)
  {
    log.trace("resolving (init part 2): " + req);

    // Check to see if we need to prime a stub zone.
    if (primeStub(event, state, req.getQName(), req.getQClass()))
    {
      // A priming request was sent.
      return false;
    }

    // most events just get forwarded to the next state.
    return nextState(event, req, state, IterEventState.INIT_REQUEST_3_STATE);
  }

  /**
   * Process the third part of the initial request handling. This state exists
   * as a separate state so that queries that generate stub priming events
   * will get the tail end of the init process but not repeat the stub priming
   * check.
   * 
   * @param event The event being processed.
   * @param req The request for that event.
   * @param state The module state for that event.
   * @return true, advancing the event to the QUERYTARGETS_STATE.
   */
  private boolean processInitRequest3(DNSEvent event, Request req,
      IterEventState state)
  {
    log.trace("resolving (init part 3): " + req);
    // If the RD flag wasn't set, then we just finish with the cached referral
    // as the response.
    if (!req.getFlag(Flags.RD))
    {
      event.setResponse(state.currentDelegationMessage);
      return finalState(event, req, state);
    }

    // After this point, unset the RD flag -- this query is going to be sent
    // to an auth. server.
    req.unsetFlag(Flags.RD);

    // Jump to the next state.
    return nextState(event, req, state, IterEventState.QUERYTARGETS_STATE);
  }

  /**
   * Given a basic query, generate a "target" query. These are subordinate
   * queries for missing delegation point target addresses.
   * 
   * @param forEvent The event triggering this query.
   * @param forState The state associated with the forEvent.
   * @param targetName The qname.
   * @param targetType The qtype (either A or AAAA)
   * @param targetClass the qclass.
   * 
   * @return A DNSEvent containing the query and property state routing.
   */
  private DNSEvent generateTargetQuery(DNSEvent forEvent,
      IterEventState forState, Name targetName, int targetType,
      int targetClass)
  {
    DNSEvent target_request = generateLocalRequest(targetName,
        targetType,
        targetClass,
        forEvent,
        IterEventState.INIT_REQUEST_STATE,
        IterEventState.TARGET_RESP_STATE);
    IterEventState s = getModuleState(target_request);
    s.currentDelegationPoint = forState.currentDelegationPoint;

    return target_request;
  }

  /**
   * Given an event at a certain state, generate zero or more target queries
   * for it's current delegation point.
   * 
   * @param event The event in question.
   * @param request The current request for the event.
   * @param state The state for the event.
   * @param ntargets The maximum number of targets to query for.
   * @return The number of queries generated and processed, which may be zero
   *         if there were no missing targets.
   */
  private int queryForTargets(DNSEvent event, Request request,
      IterEventState state, int ntargets)
  {
    int query_count = 0;

    DelegationPoint dp = state.currentDelegationPoint;
    // Generate target requests. Basically, any missing targets are queried
    // for here, regardless if it is necessary to do so to continue
    // processing.
    Name dp_name = dp.getName();
    List targetNames = dp.getMissingTargetNames();

    if (targetNames.isEmpty())
    {
      log.trace("no missing targets to query for dp: " + dp.getName());
      return 0;
    }

    int n = 0;
    for (Iterator i = targetNames.iterator(); i.hasNext();)
    {
      Name name = (Name) i.next();

      // Sanity check: if the target name is at or *below* the delegation
      // point itself, then this will be (potentially) unresolvable. This is
      // the one case where glue *must* have been present.
      // FIXME: at this point, this *may* be resolvable, so perhaps we should
      // issue the query anyway and let it fail.
      if (name.subdomain(dp_name))
      {
        log.debug("skipping target name '" + name
            + "' because it should have been glue");
        continue;
      }

      if (mSupportsIPv6)
      {
        // Send the AAAA request.
        DNSEvent target_event = generateTargetQuery(event,
            state,
            name,
            Type.AAAA,
            request.getQClass());
        processLocalRequest(target_event);
        query_count++;
      }
      // Send the A request.
      DNSEvent target_event = generateTargetQuery(event,
          state,
          name,
          Type.A,
          request.getQClass());
      processLocalRequest(target_event);
      query_count++;

      // mark this target as in progress.
      dp.markTarget(name);

      // if ntargets is negative, there is no maximum, otherwise only query
      // for ntarget names.
      if (ntargets > 0 && ++n > ntargets)
      {
        break;
      }
    }

    return query_count;
  }

  /**
   * This is the request event state where the request will be sent to one of
   * its current query targets. This state also handles issuing target lookup
   * queries for missing target IP addresses. Queries typically iterate on
   * this state, both when they are just trying different targets for a given
   * delegation point, and when they change delegation points. This state
   * roughly corresponds to RFC 1034 algorithm steps 3 and 4.
   * 
   * @param event The (request) event being processed.
   * @param request The current request for that event.
   * @param state The current state for that event.
   * 
   * @return true if the event requires more request processing immediately,
   *         false if not. This state only returns true when it is generating
   *         a SERVFAIL response because the query has hit a dead end.
   */
  private boolean processQueryTargets(DNSEvent event, Request request,
      IterEventState state)
  {
    // NOTE: a request will encounter this state for each target it needs to
    // send a query to. That is, at least one per referral, more if some
    // targets timeout or return throwaway answers.

    log.trace("processQueryTargets(" + event.getRequest() + "): "
        + state.targetQueries + ", " + state.currentQueries);

    // Make sure that we haven't run away
    // FIXME: is this check even necessary?
    if (state.referralCount > 30)
    {
      log.debug("request has exceeded the maximum number of referrrals with "
          + state.referralCount);
      return errorResponse(event, request, state, Rcode.SERVFAIL);
    }

    int tf_policy = 0;
    if (event.getDepth() < mTargetFetchPolicy.length)
    {
      tf_policy = mTargetFetchPolicy[event.getDepth()];
    }

    // if there is a policy to fetch missing targets opportunistically, do it.
    // we rely on the fact that once a query (or queries) for a missing name
    // have been issued, they will not be show up again.
    if (tf_policy != 0)
    {
      state.targetQueries = queryForTargets(event, request, state, tf_policy);
    }
    else
    {
      state.targetQueries = 0;
    }

    Name zonename = state.currentDelegationPoint.getName();
    Host[] targets = state.currentDelegationPoint.getTargets();

    // Add the current set of unused targets to our queue.
    if (state.targetQueue == null) state.targetQueue = new LinkedList();
    state.targetQueue.addAll(Arrays.asList(targets));

    // Order the current set of available targets.
    if (mServerSelection != null)
    {
      mServerSelection.orderTargets(state.targetQueue);
    }

    Host target = null;

    // Select the next usable target, filtering out unsuitable targets.
    while (!state.targetQueue.isEmpty())
    {
      target = (Host) state.targetQueue.removeFirst();

      // skip targets that are in the "do-not-query" list.
      if (mDoNotQueryAddresses != null
          && mDoNotQueryAddresses.contains(target.getAddress()))
      {
        target = null;
        continue;
      }

      // skip targets that cannot be used due to transport.
      if (!mSupportsIPv6
          && target.getAddress() instanceof java.net.Inet6Address)
      {
        target = null;
        continue;
      }

      // skip targets that are cached as being lame.
      if (mHostCache.isLame(target, zonename, request.getQClass()))
      {
        target = null;
        continue;
      }

      // TODO: skip targets that have timed out too recently?

      // Otherwise, we have a valid target.
      break;
    }

    // If no usable target was selected...
    if (target == null)
    {
      // Here we distinguish between three states: generate a new target
      // query, just wait, or quit (with a SERVFAIL).
      // We have the following information: number of active target queries,
      // number of active current queries, the presence of missing targets at
      // this delegation point, and the given query target policy.

      // Check for the wait condition. If this is true, then an action must be
      // taken.
      if (state.targetQueries == 0 && state.currentQueries == 0)
      {
        // If there is nothing to wait for, then we need to distinguish
        // between generating (a) new target query, or failing.
        if (state.currentDelegationPoint.hasMissingTargets())
        {
          log.trace("querying for next missing target");
          state.targetQueries = queryForTargets(event, request, state, 1);
        }

        // Since a target query might have been made, we need to check again.
        if (state.targetQueries == 0)
        {
          log.debug("out of query targets -- returning SERVFAIL");
          // fail -- no more targets, no more hope of targets, no hope of a
          // response.
          return errorResponse(event, request, state, Rcode.SERVFAIL);
        }
      }

      // otherwise, we have no current targets, so submerge until one of the
      // target or direct queries return.

      if (state.targetQueries > 0 && state.currentQueries > 0)
      {
        log.trace("no current targets -- waiting for targets to resolve "
            + "or outstanding queries to respond");
      }
      else if (state.targetQueries > 0)
      {
        log.trace("no current targets -- waiting for targets to resolve: "
            + state.targetQueries + " outstanding target queries");
      }
      else
      {
        log.trace("no current targets -- waiting for "
            + "outstanding queries to respond: " + state.currentQueries
            + " current queries outstanding");
      }

      return false;
    }

    // We have a valid target.

    // Set the wakeup timer -- this will normally be determined by target RTT.
    // For now, a constant.
    int wakeup_time = -1;
    if (mServerSelection != null)
    {
      wakeup_time = mServerSelection.getWakeupTimer(target);
    }

    if (wakeup_time > 0)
    {
      request.setWakeupTime(wakeup_time);
    }

    log.trace("sending query for: " + request + " to target " + target);
    request.setTarget(target);
    state.currentQueries++; // note the query.
    mNextResolver.processRequest(event);
    return false;
  }

  /**
   * Process the query response. All queries end up at this state first. This
   * process generally consists of analyzing the response and routing the
   * event to the next state (either bouncing it back to a request state, or
   * terminating the processing for this event).
   * 
   * @param event The event to process.
   * @param request The current request for the event.
   * @param state The state for the event.
   * 
   * @return true if the event requires more immediate processing, false if
   *         not. This is generally only true when forwarding the request to
   *         the final state (i.e., on answer).
   */
  private boolean processQueryResponse(DNSEvent event, Request request,
      IterEventState state)
  {
    // If the event timed out, it either actually timed out, or was just woken
    // up. In either case, send this event back to the QUERYTARGETS state for
    // the next target.
    Response response = event.getResponse();
    int resp_status = response.getResponseStatus();

    if (resp_status != Response.OK)
    {
      if (resp_status == Response.TIMEOUT)
      {
        log.debug("query response was a timeout");
        // note that this outstanding current query has finished.
        state.currentQueries--;
      }
      else
      {
        log.trace("query response was a wakeup");
      }

      // go back to the QUERYTARGETS state.
      return nextState(event,
          request,
          state,
          IterEventState.QUERYTARGETS_STATE);
    }

    // note that this outstanding current query has finished.
    state.currentQueries--;

    // sanitize before typing (since sanitization can change the type).
    sanitize(event, request, state);

    SMessage resp = response.getSMessage();
    int resptype = ResponseType.typeWireResponse(resp,
        request,
        state.currentDelegationPoint);
    response.setResponseType(resptype);

    DNSEvent event_copy;

    switch (resptype)
    {
      case ResponseType.ANSWER :
        // ANSWER type responses terminate the query algorithm, so they are
        // sent on their way.

        log.trace("query response was ANSWER");
        // FIXME: there is a question about whether this gets stored under the
        // original query or most recent query. The original query would
        // reduce cache work, but you need to apply the prependList before
        // caching, and also cache under the latest query.
        
        // mDNSCache.store(resp, event.getOrigRequest(), false);
        mDNSCache.store(resp, request, false);
        
        // close this event so that any outstanding requests will be
        // discarded.
        event.close();

        // advance this event to it's final state.
        return finalState(event, request, state);

      case ResponseType.REFERRAL :
        // REFERRAL type responses get a reset of the delegation point, and go
        // back to the QUERYTARGETS_STATE.
        log.trace("query response was REFERRAL");

        // Store the referral under the current query
        mDNSCache.store(resp, request, true);
        // Reset the event state, setting the current delegation point to the
        // referral.
        state.reset(resp, DelegationPoint.messageToDelegationPoint(resp,
            mHostCache));
        // Count this as a referral.
        state.referralCount++;

        // A copy is made of the current event so that the existing event can
        // be closed, in case there are still outstanding "currentQueries",
        // either sent or unsent. Note that the event_copy will share the same
        // module state as the original event.
        event_copy = (DNSEvent) event.clone();
        log.trace("copied event for next round -- " + event_copy);
        event.close();

        return activateEvent(event_copy,
            getModuleState(event_copy),
            IterEventState.QUERYTARGETS_STATE);

      case ResponseType.CNAME :
        // CNAME type responses get a query restart (i.e., get a reset of the
        // query state and go back to INIT_REQUEST_STATE).
        log.trace("query response was CNAME");

        // Process the CNAME response.
        Name sname = handleCNAMEResponse(state, request, resp);
        // cache the CNAME response under the current query
        mDNSCache.store(resp, request, false);

        // set the current request's qname to the new value.
        request.setQName(sname);

        // Clear the query state, since this is a query restart.
        state.reset(null, null);
        // Note the query restart.
        state.queryRestartCount++;

        // A copy is made of the current event so that the existing event can
        // be closed, in case there are still outstanding "currentQueries",
        // either sent or unsent. Note that the event_copy will share the same
        // module state object with the original.
        event_copy = (DNSEvent) event.clone();
        log.trace("copied event for next round (restart) -- " + event_copy);
        event.close();

        // Send the copy to the INIT_REQUEST_STATE.
        return activateEvent(event_copy,
            getModuleState(event_copy),
            IterEventState.INIT_REQUEST_STATE);

      case ResponseType.LAME :
        // Cache the LAMEness.
        mHostCache.markLame(response.getSource(),
            state.currentDelegationPoint.getName(),
            request.getQClass());

      case ResponseType.THROWAWAY :
        // LAME and THROWAWAY responses are handled the same way. In this
        // case, the event is just sent directly back to the
        // QUERYTARGETS_STATE without resetting anything, because, clearly,
        // the next target must be tried.
        log.debug("query response was LAME/THROWAWAY");
        break;
      default :
        log.warn("A query response came back with an unknown type: "
            + ResponseType.string(resptype) + " (" + resptype + ")");
        break;
    }

    // LAME, THROWAWAY and "unknown" all end up here.
    // Recycle to the QUERYTARGETS state to hopefully try a different target.
    return nextState(event, request, state, IterEventState.QUERYTARGETS_STATE);
  }

  /**
   * Do final processing on responses to target queries. Events reach this
   * state after the iterative resolution algorithm terminates. This state is
   * responsible for reactiving the original event, and housekeeping related
   * to received target responses (caching, updating the current delegation
   * point, etc).
   * 
   * @param event The (response) event to process.
   * @param request The current request associated with the event.
   * @param state The state associated with the event.
   * @return true if the event requires more (response) processing
   *         immediately, false if not. This particular state always returns
   *         false.
   */
  private boolean processTargetResponse(DNSEvent event, Request request,
      IterEventState state)
  {
    DNSEvent forEvent = event.forEvent();
    IterEventState forState = getModuleState(forEvent);

    // Tell the originating event that this target query has finished
    // (regardless if it succeeded or not).
    forState.targetQueries--;

    Response response = event.getResponse();
    SMessage resp = response.getSMessage();

    // Check to see if the originating event is still interested.
    if (forEvent.isClosed())
    {
      // If not (i.e., it was closed as the process moved on or finished),
      // just stop processing this event and do not reactive the forEvent.
      return false;
    }

    // This response is relevant to the current query, so we add (attempt to
    // add, anyway) this target(s) and reactivate the original event.
    // NOTE: we could only look for the AnswerRRset if the response type was
    // ANSWER.
    SRRset rrset = resp.findAnswerRRset(request.getQName(), request
        .getQType(), request.getQClass());
    if (rrset != null)
    {
      forState.currentDelegationPoint.addTargets(rrset, mHostCache);
    }
    else
    {
      forState.currentDelegationPoint.failTarget(request.getQName());
    }

    // Reactivate the forEvent, now that it has either a new target or a
    // failed target.
    return activateEvent(forEvent,
        forState,
        IterEventState.QUERYTARGETS_STATE);
  }

  /**
   * This handles the response to a priming query. This is used to handle both
   * root and stub priming responses. This is basically the equivalent of the
   * QUERY_RESP_STATE, but will not handle CNAME responses and will treat
   * REFERRALs as ANSWERS. It will also update and reactivate the originating
   * event.
   * 
   * @param event The priming event.
   * @param request The current request for that event.
   * @param state The current state for that event.
   * 
   * @return true if the event needs more immediate processing, false if not.
   *         This state always returns false.
   */
  private boolean processPrimeResponse(DNSEvent event, Request request,
      IterEventState state)
  {
    DNSEvent forEvent = event.forEvent();
    IterEventState forState = getModuleState(forEvent);

    Response response = event.getResponse();
    SMessage resp = response.getSMessage();

    DelegationPoint dp = null;

    if (response.getResponseType() == ResponseType.ANSWER)
    {
      // Convert our response to a delegation point
      dp = DelegationPoint.messageToDelegationPoint(resp, mHostCache);
    }

    if (dp == null)
    {
      // if there is no convertable delegation point, then the ANSWER type was
      // (presumably) a negative answer.
      log.debug("prime response was not a positive " + "ANSWER; failing");
      // note that this will finish the forEvent.
      return activateErrorResponse(forEvent, forState, Rcode.SERVFAIL);
    }

    log.debug("priming successful for " + request.getQName() + "/"
        + DClass.string(request.getQClass()));

    // This event is finished.
    event.close();

    forState.currentDelegationMessage = resp;
    forState.currentDelegationPoint = dp;

    // root priming responses go to init stage 2, priming stub responses to to
    // stage 3.
    return activateEvent(forEvent, forState, state.primingStub
        ? IterEventState.INIT_REQUEST_3_STATE
        : IterEventState.INIT_REQUEST_2_STATE);
  }

  /**
   * This handles the final state for first-tier responses (i.e., responses to
   * externally generated queries).
   * 
   * @param event The response event.
   * @param request The current request for this event.
   * @param state The current state for this event.
   * @return true if the event needs more processing, false if not. Since this
   *         is the final state for an event, it always returns false.
   */
  private boolean processFinished(DNSEvent event, Request request,
      IterEventState state)
  {
    log.trace("finishing processing for " + request);

    Response response = event.getResponse();
    SMessage resp = response.getSMessage();

    // Make sure the final response contains the original question.
    resp.setQuestion(event.getOrigRequest().getQuestion());

    // Make sure that the RA flag is set (since the presence of this module
    // means that recursion is available)
    resp.getHeader().setFlag(Flags.RA);

    // Clear the AA flag
    // FIXME: does this action go here or in some other module?
    resp.getHeader().unsetFlag(Flags.AA);

    // If we have stuff to prepend, prepend it.
    if (resp.getRcode() != Rcode.SERVFAIL && state.prependList != null
        && state.prependList.size() > 0)
    {
      resp.prependRRsets(state.prependList, Section.ANSWER);
    }

    // If we are using a private TTL, trim the response.
    if (mPrivateTTL > 0)
    {
      IterUtils.setPrivateTTL(resp, mPrivateTTL);
    }

    // Send the response along.
    mPrevResolver.processResponse(event);
    return false;
  }

}
